#!‪C:\Python38\python.exe
import xml.etree.ElementTree as ET
import sys
import time

def exception_handle_decorator(func):
    def inner(processor_instance, element, ind, short, braced=False):
        try:
            func(processor_instance, element, ind, short, braced)
        except Exception as ex:
            type, value, traceback = sys.exc_info()          
            put(ind + "###ERROR### " + ex.__class__.__name__ + "(" + str(ex) + ") parsing " +func.__name__)
            if rethrow:
                raise
    return inner

class FailedToFind(Exception):
    
    def __init__(self, message="failed to find something"):
        self.message = message
        super().__init__(self.message)
    
# ========= PARSER CLASS ==============
class parser(object):

    def find_attrib(self, element, attrib_name):
        #print(len(element.attrib))
        for key in element.attrib:
            #print(key)
            if str(key)== attrib_name:
                return element.attrib[key]

    def find_attrib_ending(self, element, attrib_name):
        for key in element.attrib:
            if str(key).endswith(attrib_name):
                return element.attrib[key]

    def find_tags_ending(self, element, target):
        for child in list(element):
            childstr = str(child.tag)
            if childstr.endswith(target):
                yield child

    def find(self, element, *targets):
        if targets is None or len(targets) == 0:
            return element
            
        e = element
        for target in targets:
            
            # find content in tag special-cases
            if target.startswith('ATT:'):
                #print(target[4:])
                #print (e.attrib)
                return self.find_attrib(e, target[4:])
            
            if target.startswith('ATT_ENDING:'):
                #print(element)
                return self.find_attrib_ending(e, target[11:])

            if target == 'TEXT:':
                return e.text

            # search for tag    
            for child in list(e):
                childstr = str(child.tag)
                if childstr.endswith(target):
                    #print('found:'+target)
                    e = child
                    break
        if e == element:
            raise FailedToFind( "Failed to find:" + "|".join(targets))
        return e

    def find_or_none(self, element, *targets):
        try:
            return self.find(element, *targets);
        except FailedToFind:
            return None
        
    def multifind(self, element, target):
        for child in list(element):
            childstr = str(child.tag)
            if childstr.endswith(target):
                yield child
                
# ======================================================================
# ======================================================================
# ======================== PROCESSOR CLASS =============================
# ======================================================================
# ======================================================================

class processor(object):

    def __init__(self):
        self.dispatch={
            'Sequence':                 self.Sequence,
            'TryCatch':                 self.TryCatch,
            'Throw':                    self.Throw,
            'MultipleAssign':           self.MultipleAssign,
            'Click':                    self.Click,
            'OnUiElementAppear':        self.OnUiElementAppear,
            'WaitUiElementAppear':      self.WaitUiElementAppear,
            'Delay':                    self.Delay,
            '}Assign':                  self.Assign,
            'LogMessage':               self.LogMessage,
            'Comment':                  self.Comment,
            'UiElementExists':          self.UiElementExists,
            'CheckTrue':                self.CheckTrue,
            'CheckFalse':               self.CheckFalse,
            'GetValue':                 self.GetValue,
            'Switch':                   self.Switch,
            'Switch.Default':           self.SwitchDefault,
            'InvokeWorkflowFile':       self.InvokeWorkflowFile,
            'InvokeCode':               self.InvokeCode,
            'LookupDataTable':          self.LookupDataTable,
            '}If':                      self.If,
            'TypeInto':                 self.TypeInto,
            'DoWhile':                  self.DoWhile,
            'CommentOut':               self.CommentOut,
            'GetRobotAsset':            self.GetRobotAsset,
            'OpenBrowser':              self.OpenBrowser,
            'GetAttribute':             self.GetAttribute,
            'ElementScope':             self.ElementScope,
            'ExtractData':              self.ExtractData,
            'ForEach':                  self.ForEach,
            'ForEachRow':               self.ForEachRow,
            'Replace':                  self.Replace,
            'BrowserScope':             self.BrowserScope,
            'CloseTab':                 self.CloseTab,
            'Rethrow':                  self.Rethrow,
            'Flowchart':                self.Flowchart,
            'FlowStep':                 self.FlowStep,
            'FlowStep.Next':            self.FlowStepNext,
            'AddDataRow':               self.AddDataRow,
            'BuildDataTable':           self.BuildDataTable,
            'WindowScope':              self.WindowScope,
            'MaximizeWindow':           self.MaximizeWindow,
            'SendHotkey':               self.SendHotkey,
            'ClickImage':               self.ClickImage,
            'While':                    self.While,
            'GetQueueItems':            self.GetQueueItems,
            'DeleteQueueItems':         self.DeleteQueueItems,
            'AddToCollection':          self.AddToCollection,
            'ShouldStop':               self.ShouldStop,
            'CloseWindow':              self.CloseWindow,
            'GetRobotCredential':       self.GetRobotCredential,
            'RetryScope':               self.RetryScope,
            'TypeSecureText':           self.TypeSecureText,
            'Check':                    self.Check
            
           # 'ActivityAction.Argument':  self.Argument
        }
        
        self.hash_selector = True
        self.hash_datatable = True

        self.ignore=[
            'ViewState',
            'Variables',
            'Argument',
            'Reference',
            'DoWhile.Condition',
            'While.Condition',
            '}Members',
            'VisualBasic.Settings',
            'VirtualizedContainerService.HintSize',
            'WorkflowViewState.IdRef',
            'NamespacesForImplementation',
            'ReferencesForImplementation',
            'Flowchart.StartNode'
        ]

    def start(self, element, ind, short, hash_selector, show_datatable):
        self.hash_selector = hash_selector
        if show_datatable:
            self.hash_datatable = False

        if str(element.tag).endswith('}Activity'):
            self.process_activity(element, ind, short)
        
    def process_activity(self, element, ind, short):
        child, handler = self.find_child(element)
        handler(child, '' , short)

    # handy methods

    def indent(self, indent):
        return indent + '    '

    def show_arg(self, text):
        if len(text) == 0:
            return ''
        if text[0] == '[':
            return text[1:-1]
        else:
            return text

    def parse_text_or_literal(self, element):
        text = find(element, 'ATT:Text')
        if not text:
            textchild = find(element, 'Text')
            inarg = find(textchild, 'InArgument')
            lit = find(inarg, 'Literal')
            text = find(lit, 'ATT:Value')
        return text

    def escape(self, text):
        text = text.replace('&lt;','<')
        text = text.replace('&gt;','>')
        text = text.replace('&quot;','"')
        return text

    def ends_with_any(self, text, *ends):
        for end in ends:
            if text.endswith(end):
                return True
        return False
        
    def find_child(self, element):
        return next(self.find_children(element))

    def find_children(self, element):
        for child in list(element):
            childstr = str(child.tag)
            found = False
            for key in list(self.dispatch):
                if childstr.endswith(key):
                    handler = self.dispatch[key]
                    yield(child, handler)
                    found = True
                    break
                
            # Highlight when child is not recognised. Suppress a few outliers
            if not found and not self.ends_with_any(childstr, *self.ignore):
                put('------------------------' + childstr)
           

    def find_children_of_type(self, element, target):
        for child in list(element):
            childstr = str(child.tag)
            if childstr.endswith(target):
                yield child
                               
    def display_name(self, element, ind, skip_if_matches=""):
        name = find(element,'ATT:DisplayName')
        if name:
            if not name == skip_if_matches:
                put(ind + 'Name: ' + name)

    def display_child(self, element, ind, short, *path):
        ind2 = self.indent(ind)
        body = find(element,*path)
        child, handler = self.find_child(body)
        handler(child, ind2, short, braced=True)
        put(ind + '}')

    def display_children(self, element, ind, short, *path):
        ind2 = self.indent(ind)
        body = find(element, *path)
        for child, handler in self.find_children(body):
            handler(child, ind2, short, True)  
        put(ind + '}')

    
    def display_variables(self, element, ind, short):
        variable_tag = find_or_none(element,'.Variables')
        if variable_tag:
            variables = multifind(variable_tag, 'Variable')
            for variable in variables:
                name = find(variable,'ATT:Name')
                var_type = find(variable,'ATT_ENDING:TypeArguments')
                if var_type.startswith('x:'):
                    var_type = var_type[2:]
                default = find(variable,'ATT:Default')
                if default:
                    put(ind + var_type + ' ' + name + ' = ' + default)
                else:
                    put(ind + var_type + ' ' + name)    
        
    def display_selector (self, selector, ind):
        i = ind # Make a copy!
        if self.hash_selector:
            put(ind + 'Selector: hash=' + str(hash(selector)))
            return
        selector = selector.replace('\r','')
        selector = selector.replace('\n','')
        selector = selector.replace('&lt;','<')
        selector = selector.replace('&gt;','>')
        selector = selector.replace(';&#xA;','')
        selector = selector.replace('#xA;','')
        selector = selector.replace('><','>linebreak-jkjdiea<')
        selectors = selector.split('linebreak-jkjdiea')
        first = True
        for selector in selectors:
            if first:
                put(i + 'Selector: ' + selector)
                first = False
            else:
                put(i + '          ' + selector)
                pass        

    # ----------- sequence ---------
    @exception_handle_decorator
    def Sequence(self, element, ind, short, braced=False):
        ind2 = self.indent(ind)
        if short:
            if braced:
                self.display_variables(element, ind, short)
                # render children
                for child, handler in self.find_children(element):
                    handler(child, ind, short, True)
            else:
                put(ind + '{')
                self.display_variables(element, ind2, short)
                self.display_children(element, ind, short)
        else:
            put(ind + 'Sequence(')
            self.display_name(element, ind2, 'Sequence')
            put(ind + '){')
            self.display_variables(element, ind2, short)
            self.display_children(element, ind, short)

    #------- try-catch ---------------
    @exception_handle_decorator
    def TryCatch(self, element, ind, short, braced=False):
        ind2 = self.indent(ind)
        
        # render open
        if short:
            put(ind + 'try{')
        else:       
            put(ind + 'try(')
            self.display_name(element, ind)
            put(ind + '){')
        # render inside try
        self.display_variables(element, ind2, short)
        self.display_children(element, ind, short, 'TryCatch.Try','WindowScope','WindowScope.Body','ActivityAction')

        # render catches
        catches = find(element,'TryCatch.Catches')
        for catch in self.find_children_of_type(catches, 'Catch'):
            deltype = find(catch, 'ActivityAction','ActivityAction.Argument','DelegateInArgument','ATT_ENDING:TypeArguments')
            delname = find(catch, 'ActivityAction','ActivityAction.Argument','DelegateInArgument','ATT_ENDING:Name')
            put(ind + "catch("+deltype+" "+delname+"){")
            self.display_children(catch, ind, short,'ActivityAction')
            
        final = find_or_none(element,'TryCatch.Finally')
        if final:
            put(ind + 'finally{')
            self.display_children(final, ind, short)

            
            #act = self.rel(catch,'ActivityAction')
            #delig = self.rel(act,'ActivityAction.Argument','DelegateInArgument')
            #deltype = self.find_attrib_ending('TypeArguments', delig.attrib)
           # delname = self.find_attrib('Name', delig.attrib)
            #print(ind + "catch("+deltype+" "+delname+"){")
            #child, handler = self.find_child(act)
            #handler(child, ind2, short, True)
            #print(ind + "}")
                
            


            #print('-------------')
            #for key in delig.attrib:
            #    print(ind + str(key) + ':')
            #    print(ind + delig.attrib[key])

    # --------- Throw ----------
    @exception_handle_decorator
    def Throw(self, element, ind, short, braced=False):
        exception = find(element,'ATT:Exception')
        exception = self.show_arg(exception)
        put(ind + 'throw ' + exception)


    # --------- ui:Click --------
    @exception_handle_decorator
    def Click(self, element, ind, short, braced=False):
        ind2 = self.indent(ind)
        clicktype = find(element,'ATT:ClickType')
        #print(find(element,'Target','Target'))
        selector = find(element,'Target','Target','ATT:Selector')
        put(ind + clicktype.lower() + '(')
        self.display_name(element, ind2)
        self.display_selector (selector, ind2)
        put(ind + ')')
        
    # --------- ui:MultipleAssign ---------------
    @exception_handle_decorator
    def MultipleAssign(self, element, ind, short, braced=False):
        ind2 = self.indent(ind)
        assignlist = find(element,'AssignOperations','List')
        for assign in multifind(assignlist, 'AssignOperation'):
            ass_to = find(assign,'To','OutArgument','TEXT:')
            ass_from = find(assign,'Value','InArgument','TEXT:')
            put(ind + self.show_arg(ass_to)+ ' = ' + self.show_arg(ass_from))

    # --------- ui:OnUiElementAppear ------------
    @exception_handle_decorator
    def OnUiElementAppear(self, element, ind, short, braced=False):
        ind2 = self.indent(ind)
        selector = find(element,'Target','Target','ATT:Selector')
        put(ind + 'OnUiElementAppear(')
        self.display_name(element, ind2)
        self.display_selector (selector, ind2)
        put(ind + '){')
        self.display_variables(element, ind, short)
        self.display_children(element, ind, short, 'Body')
        
    # --------- WaitUiElementAppear -------------
    @exception_handle_decorator
    def WaitUiElementAppear(self, element, ind, short, braced=False):
        ind2 = self.indent(ind)
        selector = find(element,'Target','Target','ATT:Selector')
        put(ind + 'WaitUiElementAppear(')
        # ----- had to do something odd here - other selectors might do the same ---
        if selector:
            self.display_selector (selector, ind2)
        else:
            selector = find(element, 'Target','Target','Selector','InArgument','Literal','TEXT:')
            if selector:
                self.display_selector (selector, ind2)
            else:
                put(ind2 + 'FAILED TO PARSE SELECTOR')             
        put(ind + ')')
        
    # ---------- Delay --------------------------
    @exception_handle_decorator
    def Delay(self, element, ind, short, braced=False):
        ind2 = self.indent(ind)
        duration = find(element,'ATT:Duration')
        put(ind + 'Delay( Duration: ' + duration)
        self.display_name(element, ind2, 'Delay')
        put(ind + ')')
        
    # -------- Assign ------------------
    @exception_handle_decorator
    def Assign(self, element, ind, short, braced=False):
        ass_to = find(element,'To','OutArgument','TEXT:')
        ass_from = find(element,'Value','InArgument','TEXT:')
        put(ind + self.show_arg(ass_to)+ ' = ' + self.show_arg(ass_from))
        
    # -------- LogMessage --------------
    @exception_handle_decorator
    def LogMessage(self, element, ind, short, braced=False):
        ind2 = self.indent(ind)
        level = find(element,'ATT:Level')
        message = find(element,'ATT:Message')
        message = message.replace('&quot;','"')      
        message = self.show_arg(message)
        put(ind + 'LogMessage(')
        put(ind2 + 'Level : ' + level)
        put(ind2 + 'Message: ' + message)
        put(ind + ')')
        
    # -------- Comment --------------
    @exception_handle_decorator
    def Comment(self, element, ind, short, braced=False):
        message = find(element,'ATT:Text')
        put(ind + '// ' + message)
        
    # -------- UiElementExists ------
    @exception_handle_decorator
    def UiElementExists(self, element, ind, short, braced=False):
        ind2 = self.indent(ind)
        selector = find(element,'Target','Target','ATT:Selector')
        put(ind + 'UiElementExists(')
        self.display_name(element, ind, 'Element Exists')
        self.display_selector (selector, ind2)
        put(ind + ')')

    #----- CheckTrue --------------
    @exception_handle_decorator
    def CheckTrue(self, element, ind, short, braced=False):
        ind2 = self.indent(ind)
        expression = self.show_arg(find(element,'ATT:Expression'))
        put(ind + 'CheckTrue(')
        self.display_name(element, ind2, 'Check True')
        put(ind2 + 'Expression: ' + expression)
        put(ind + ')')

    # ----------- CheckFalse ---------
    @exception_handle_decorator
    def CheckFalse(self, element, ind, short, braced=False):
        ind2 = self.indent(ind)
        expression = self.show_arg(find(element,'ATT:Expression'))
        put(ind + 'CheckFalse(')
        self.display_name(element, ind, 'Check False')
        put(ind2 + 'Expression: ' + expression)
        put(ind + ')')
        
    #------ GetValue -------------
    @exception_handle_decorator
    def GetValue(self, element, ind, short, braced=False):
        ind2 = self.indent(ind)
        selector = find(element,'Target','Target','ATT:Selector')
        out_argument = self.show_arg(find(element,'Value','OutArgument','TEXT:'))
        put(ind + 'GetValue(')
        self.display_name(element, ind2, 'Get Value')
        self.display_selector (selector, ind2)
        put(ind2 + 'Out: ' + out_argument)
        put(ind + ')')
        
    # ------- Switch -----------
    @exception_handle_decorator
    def Switch(self, element, ind, short, braced=False):
        ind2 = self.indent(ind)
        expression = self.show_arg(find(element,'ATT:Expression'))
        put(ind + 'Switch(')
        self.display_name(element, ind2, 'Switch')
        put(ind2 + 'Expression: ' + expression)
        put(ind + '){')
        self.display_variables(element, ind, short)
    
        for case, handler in self.find_children(element): # ----------- NEEDS TO BE LESS SPECIFIC ----------
            if str(case.tag).endswith('Switch.Default'):
                put(ind2 + 'case DEFAULT:')
                case, handler = self.find_child(case)
                handler(case, ind2, short, False)
            else:
                key = find(case,'ATT_ENDING:Key')
                put(ind2 + 'Case: ' + key)
                ind3 = self.indent(ind2)            
                put(ind2 + '{')             
                handler(case, ind3, short, False)
                put(ind2 + '}')            
        put(ind + '} // end of switch')

    # ------- SwitchDefault -----------
    @exception_handle_decorator
    def SwitchDefault(self, element, ind, short, braced=False):
        pass
        ind2 = self.indent(ind)
        put(ind2 + 'case DEFAULT:')
        #case, handler = self.find_child(element)
        #handler(case, ind2, short, False)

    # ------------ InvokeWorkflowFile -----------
    @exception_handle_decorator
    def InvokeWorkflowFile(self, element, ind, short, braced=False):
        ind2 = self.indent(ind)
        filename = find(element,'ATT:WorkflowFileName')
        put(ind + 'InvokeWorkflowFile(')
        self.display_name(element, ind2, 'Invoke Workflow')
        put(ind2 + 'File: ' + filename)
        put(ind2 + '----- ')

        
        # display arguments
        args = find(element, 'Arguments')
        for arg in self.find_children_of_type(args, 'Argument'):
            key = find(arg, 'ATT_ENDING:Key')
            text = self.show_arg(find(arg, 'TEXT:'))
            if arg.tag.endswith('InOutArgument'):
                put(ind2 + text + ' <=> ' + key)
                continue;
            if arg.tag.endswith('InArgument'):
                put(ind2 + text + ' ==> ' + key)
                continue;
            if arg.tag.endswith('OutArgument'):
                put(ind2 + text + ' <== ' + key)
                continue;               
        put(ind + ')')

    # ------------ InvokeCode -----------
    @exception_handle_decorator
    def InvokeCode(self, element, ind, short, braced=False):
        ind2 = self.indent(ind)
        code = find(element,'ATT:Code')
        put(ind + 'InvokeCode(')
        self.display_name(element, ind2, 'Invoke Code')                 
        if self.hash_selector:
            code = 'hash=' + str(hash(code))
        else:
            code = self.escape(code)
        put(ind2 + 'Code: ' + code)
        put(ind2 + '---------')
     
        # display arguments
        args = find(element, 'Arguments')
        for arg in self.find_children_of_type(args, 'Argument'):
            key = find(arg, 'ATT_ENDING:Key')
            text = self.show_arg(find(arg, 'TEXT:'))
            if arg.tag.endswith('InOutArgument'):
                put(ind2 + text + ' <=> ' + key)
                continue;
            if arg.tag.endswith('InArgument'):
                put(ind2 + text + ' ==> ' + key)
                continue;
            if arg.tag.endswith('OutArgument'):
                put(ind2 + text + ' <== ' + key)
                continue;               
        put(ind + ')')  

    # ------------ LookupDataTable ------------
    @exception_handle_decorator
    def LookupDataTable(self, element, ind, short, braced=False):
        ind2 = self.indent(ind)
        name = find(element,'ATT:DisplayName')
        lookup_column = find(element, 'ATT:LookupColumnName')
        lookup_value = self.show_arg(find(element, 'ATT:LookupValue'))
        lookup_target_column = find(element, 'ATT:TargetColumnName')
        lookup_datatable = find(element, 'ATT:DataTable')
        outp = self.show_arg(find(element, 'CellValue','OutArgument','TEXT:'))

        put(ind + 'LookupDataTable(')
        put(ind2 + 'Name: ' + name)
        put(ind2 + lookup_value + ' => ' + lookup_datatable + '[' + lookup_column +']')
        put(ind2 + outp + ' <= ' + lookup_datatable + '[' + lookup_target_column +']')
        put(ind + ')')

    # ------------ If --------------------------
    @exception_handle_decorator
    def If(self, element, ind, short, braced=False):
        ind2 = self.indent(ind)
        condition = self.show_arg(find(element, 'ATT:Condition'))
        thentag = find_or_none(element, 'Then')
        elsetag = find_or_none(element, 'Else')
        put(ind + 'If (' + condition + ')')

        has_variables = find_or_none(element,'.Variables')
        if has_variables:
            put(ind + 'variables encapsulated within the if:')
            self.display_variables(element, ind, short)
            put(ind + '')
        
        # render then - can be missing
        if thentag:
            try:
                then_child, handler = self.find_child(thentag)
            except Exception:
                put('NO RECOGNISED CHILD IN THEN')
                return

            put(ind + '{')
            handler(then_child, ind2, short, braced=True)
            put(ind + '}')
        else:
            put(ind + '{}')

        # render else - can be missing
        if elsetag:
            try:
                else_child, handler = self.find_child(elsetag)
            except Exception:
                put('NO RECOGNISED CHILD IN ELSE')
                return
            put(ind + 'else')
            put(ind + '{')
            handler(else_child, ind2, short, braced=True)
            put(ind + '}')

    # ----------- TypeInto -------------------------
    @exception_handle_decorator
    def TypeInto(self, element, ind, short, braced=False):
        ind2 = self.indent(ind)
        text = self.parse_text_or_literal(element) # Other Activities may need this where a literal is provided instead of a text attribute
        text = self.show_arg(text)
        
        selector = find(element,'Target','Target','ATT:Selector')
        put(ind + 'TypeInto(')
        self.display_name(element, ind2,'Type Into')
        self.display_selector (selector, ind2)
        put(ind2 + 'Text: ' + text)
        put(ind + ')')

    # ----------- DoWhile -----------------------
    @exception_handle_decorator
    def DoWhile(self, element, ind, short, braced=False):
        ind2 = self.indent(ind)
        condition = find_or_none(element,'Condition','VisualBasicValue','ATT:ExpressionText')
        
        if not condition:
            condition = self.show_arg(find(element,'Condition','TEXT:'))
            
        put(ind + 'DoWhile(')
        self.display_name(element, ind2,'Do While')
        put(ind2 + 'Condition: ' + condition)
                    
        # render contents
        put(ind + '){')
        self.display_variables(element, ind2, short)
        self.display_child(element, ind, short)


    # ----------- While -----------------------
    @exception_handle_decorator
    def While(self, element, ind, short, braced=False):
        ind2 = self.indent(ind)
        condition = find_or_none(element,'Condition','VisualBasicValue','ATT:ExpressionText')
        
        if not condition:
            condition = self.show_arg(find(element,'Condition','TEXT:'))
        
        put(ind + 'While(')
        self.display_name(element, ind2,'While')
        put(ind2 + 'Condition: ' + condition)
        put(ind + '){')
        self.display_variables(element, ind, short)
        self.display_child(element, ind, short)

    # ---------- CommentOut --------------------
    @exception_handle_decorator
    def CommentOut(self, element, ind, short, braced=False):
        ind2 = self.indent(ind)
        put(ind + 'CommentOut(')
        self.display_name(element, ind2, 'Comment Out')
        put(ind + '){')
        self.display_child(element, ind, short, 'Body')

    # ----------- GetRobotAsset ------------------
    @exception_handle_decorator
    def GetRobotAsset(self, element, ind, short, braced=False):
        ind2 = self.indent(ind)
        name = find(element,'ATT:DisplayName')
        assetname = self.escape(find(element,'ATT:AssetName'))
        out = self.show_arg(find(element,'Value','OutArgument','TEXT:'))
        put(ind + 'GetRobotAsset(')
        put(ind2 + 'Name: ' + name)
        put(ind2 + 'AssetName: ' + assetname)
        put(ind2 + 'Out: ' + out)
        put(ind + ')')

    # -------- Open browser ----------------
    @exception_handle_decorator
    def OpenBrowser(self, element, ind, short, braced=False):
        ind2 = self.indent(ind)
        browsertype = find(element,'ATT:BrowserType')
        url = self.show_arg(find(element, 'ATT:Url'))
        uibrowser = find_or_none(element, 'ATT:UiBrowser')
        put(ind + 'OpenBrowser(')
        self.display_name(element, ind2, 'Open Browser')
        put(ind2 + 'Url: ' + url)
        put(ind2 + 'BrowserType: ' + browsertype)
        if uibrowser:
            uibrowser = self.show_arg(uibrowser)
            put(ind2 + 'UiBrowser: ' + uibrowser)
        put(ind + '){')
        self.display_variables(element, ind, short)
        self.display_child(element, ind, short, 'Body','ActivityAction')


    # ---------- GetAttribute --------------
    @exception_handle_decorator
    def GetAttribute(self, element, ind, short, braced=False):
        ind2 = self.indent(ind)
        selector = find(element,'Target','Target','ATT:Selector')           
        out = self.show_arg(find(element,'Result','OutArgument','TEXT:'))        
        
        put(ind + 'GetAttribute(')
        self.display_name(element, ind2, 'Get Attribute')
        self.display_selector (selector, ind2)
        put(ind2 + 'Out: ' + out)
        put(ind + ')')

    # --------------- ElementScope --------------
    @exception_handle_decorator
    def ElementScope(self, element, ind, short, braced=False):
        ind2 = self.indent(ind)
        selector = find(element,'Target','Target','ATT:Selector')

        
        body = find(element,'Body','ActivityAction')
        child, handler = self.find_child(body)
        
        put(ind + 'ElementScope(')
        self.display_name(element, ind2, 'Element Scope')
        self.display_selector (selector, ind2)
        put(ind + '){')
        self.display_variables(element, ind, short)
        handler(child, ind2, short, braced=True)
        put(ind + '}')

    # ---------- ExtractData -----------------
    @exception_handle_decorator
    def ExtractData(self, element, ind, short, braced=False):
        ind2 = self.indent(ind)
        selector = find(element,'Target','Target','ATT:Selector')
        out = self.show_arg(find(element,'ATT:DataTable'))
        extract_metadata = self.escape(find(element,'ATT:ExtractMetadata'))

        put(ind + 'ExtractData(')
        self.display_name(element, ind2, 'Extract Data')
        self.display_selector (selector, ind2)
        put(ind2 + 'Out: ' + out)
        put(ind2 + 'Metadata: ' + extract_metadata)
        put(ind + ')')

    # ----------- ForEach ------------------
    @exception_handle_decorator
    def ForEach(self, element, ind, short, braced=False):
        ind2 = self.indent(ind)
        collection = self.show_arg(find(element,'ATT:Values'))
        delegate = find(element,'Body','ActivityAction','Argument','DelegateInArgument','ATT:Name')

        put(ind + 'ForEach(')
        self.display_name(element, ind2, 'For Each')
        put(ind2 + delegate + ' in ' + collection)
        has_variables = find_or_none(element,'.Variables')
        if has_variables:
            put(ind + 'variables encapsulated within the ForEach:')
            self.display_variables(element, ind, short)
            put(ind + '')
            
        put(ind + '){')
        self.display_child(element, ind, short,'Body','ActivityAction')

    # ----------- ForEachRow ------------------
    @exception_handle_decorator
    def ForEachRow(self, element, ind, short, braced=False):
        ind2 = self.indent(ind)
        datatable = self.show_arg(find(element,'ATT:DataTable'))
        delegate = find(element,'Body','ActivityAction','Argument','DelegateInArgument','ATT:Name')
        
        put(ind + 'ForEachRow(')
        self.display_name(element, ind2, 'For Each Row')
        put(ind2 + delegate + ' in datatable ' + datatable)
        put(ind + '){')
        self.display_child(element, ind, short,'Body','ActivityAction')

    # ------------ Replace ----------------------
    @exception_handle_decorator
    def Replace(self, element, ind, short, braced=False):
        ind2 = self.indent(ind)
        inputval = self.show_arg(find(element,'ATT:Input'))
        out = self.show_arg(find(element,'ATT:Result'))
        pattern = self.show_arg(find(element,'ATT:Pattern'))
        put(ind + 'Replace(')
        self.display_name(element, ind2, 'Replace')
        put(ind2 + 'Input: ' + inputval)
        put(ind2 + 'Out: ' + out)
        put(ind2 + 'Pattern: ' + pattern)
        put(ind + ')')
        
    # ------------ BrowserScope ----------------------
    @exception_handle_decorator
    def BrowserScope(self, element, ind, short, braced=False):
        ind2 = self.indent(ind)
        browser = self.show_arg(find(element,'ATT:Browser'))
        put(ind + 'BrowserScope(')
        self.display_name(element, ind2, 'Browser Scope')
        put(ind2 + 'Browser: ' + browser)
        put(ind + '){')
        self.display_variables(element, ind, short)
        self.display_child(element, ind, short, 'Body','ActivityAction')

    # --------- close tab --------
    @exception_handle_decorator
    def CloseTab(self, element, ind, short, braced=False):
        ind2 = self.indent(ind)
        browser = self.show_arg(find(element,'ATT:Browser'))       
        put(ind + 'CloseTab(')
        self.display_name(element, ind2, 'Close Tab')
        put(ind2 + 'Browser: ' + browser)
        put(ind + ')')

    # --------- close tab --------
    @exception_handle_decorator
    def Rethrow(self, element, ind, short, braced=False):
        put(ind + 'Rethrow()')

   # --------- Flowchart --------
    @exception_handle_decorator
    def Flowchart(self, element, ind, short, braced=False):
        ind2 = self.indent(ind)
        put(ind + 'Flowchart{')
        self.display_children(element, ind2, short)
        put(ind + '}')

   # --------- FlowStep --------
    @exception_handle_decorator
    def FlowStep(self, element, ind, short, braced=False):
        ind = ""
        ind2 = self.indent(ind)
        
        put(ind + '===========================================')
        put(ind + 'FlowStep{')
        #self.display_children(element, ind, short)
        #self.display_children(element, ind[:-4], short)

        ind2 = self.indent(ind)
        body = find(element)
        for child, handler in self.find_children(body):
            handler(child, ind2, short, True)
        put(ind + '}')           
    # --------- FlowStepNext --------
    @exception_handle_decorator
    def FlowStepNext(self, element, ind, short, braced=False):
        ind2 = self.indent(ind)
        #self.display_children(element, ind[:-4], short)
        for child, handler in self.find_children(element):
            handler(child, ind2, short, True)

    # ---------- AddDataRow --------
    @exception_handle_decorator
    def AddDataRow(self, element, ind, short, braced=False):
        ind2 = self.indent(ind)
        datatable = self.show_arg(find(element,'ATT:DataTable'))
        arrayrow = self.show_arg(find_or_none(element,'ATT:DataTable'))
        if arrayrow:
            arrayrow = self.escape(arrayrow)
        put(ind + 'AddDataRow(')
        self.display_name(element, ind2, 'Add Data Row')
        put(ind2 + 'DataTable: ' + datatable)
        put(ind2 + 'ArrayRow: ' + str(arrayrow))
        put(ind + ')')

    # --------- BuildDataTable --------
    @exception_handle_decorator
    def BuildDataTable(self, element, ind, short, braced=False):     
        ind2 = self.indent(ind)
        datatable = self.show_arg(find(element,'ATT:DataTable'))

        put(ind + 'BuildDataTable(')
        self.display_name(element, ind2, 'Build Datatable')
        put(ind2 + 'DataTable: ' + datatable)
        
        tableinfo = self.escape(find(element,'ATT:TableInfo'))
        if self.hash_datatable:
            tableinfo = 'hash=' + str(hash(tableinfo))
            put(ind2 + 'TableInfo: ' + tableinfo)
        else:
            put(ind2 + 'TableInfo:')
            ind3 = self.indent(ind2)
            lines = tableinfo.split('\n')
            for line in lines:
                put(ind3 + line)
        put(ind + ')')

    # ------------ WindowScope --------------
    @exception_handle_decorator
    def WindowScope(self, element, ind, short, braced=False):
        ind2 = self.indent(ind)
        selector = find(element,'ATT:Selector')
        applicationwindow = self.show_arg(find(element,'ATT:ApplicationWindow'))
        put(ind + 'WindowScope(')               
        self.display_name(element, ind2, 'Window Scope')
        self.display_selector (selector, ind2)
        put(ind2 + 'Application Window: ' + applicationwindow)
        put(ind + '){')
        self.display_variables(element, ind, short)
        self.display_child(element,ind2, short, 'Body','ActivityAction')
        
    # ------------ MaximiseWindow --------------
    @exception_handle_decorator
    def MaximizeWindow(self, element, ind, short, braced=False):
        ind2 = self.indent(ind)
        selector = find(element,'Target','Target','ATT:Selector')
        if not selector:
            selector = find(element, 'ATT:Window')
        put(ind + 'MaximiseWindow(')               
        self.display_name(element, ind2, 'Maximize Window')
        selector = find(element,'Target','Target','ATT:Selector')
        if selector:
            self.display_selector (selector, ind2)
        else:
            window = self.show_arg(find(element, 'ATT:Window'))
            put(ind2 + 'Window: ' + window)
            
        put(ind + ')')        
        
    # ------------ SendHotkey --------------
    @exception_handle_decorator
    def SendHotkey(self, element, ind, short, braced=False):
        ind2 = self.indent(ind)
        key = find(element,'ATT:Key')
        selector = find(element,'Target','Target','ATT:Selector')
        put(ind + 'SendHotkey(')               
        self.display_name(element, ind2, 'Send Hotkey')
        put(ind2 + 'Key: ' + str(key))
        self.display_selector (selector, ind2)
        put(ind + ')')

    # ------------ ClickImage --------------
    @exception_handle_decorator
    def ClickImage(self, element, ind, short, braced=False):
        ind2 = self.indent(ind)
        key = find(element,'ATT:Key')
        selector = find(element,'Target','Target','ATT:Selector')
        imghash = str(hash(find(element,'Image','ImageTarget','ATT:TargetImageBase64')))
        put(ind + 'ClickImage(')               
        self.display_name(element, ind2, 'Click Image')
        put(ind2 + 'Image: hash=' + imghash)
        self.display_selector (selector, ind2)
        put(ind + ')')

    # ------------ GetQueueItems -------------
    @exception_handle_decorator
    def GetQueueItems(self, element, ind, short, braced=False):
        ind2 = self.indent(ind)
        queuename = self.show_arg(find(element, 'ATT:QueueName'))
        queuename = self.escape(queuename)
        result  = self.show_arg(find(element, 'ATT:Result'))
        
        put(ind + 'GetQueueItems(')               
        self.display_name(element, ind2, 'Get Queue Items')
        put(ind2 + 'Queue Name: ' + queuename)
        put(ind2 + 'Out: ' + result)
        put(ind +')')
        
     # ------------ GetQueueItems -------------
    @exception_handle_decorator
    def DeleteQueueItems(self, element, ind, short, braced=False):
        ind2 = self.indent(ind)
        queueitems = self.show_arg(find(element, 'ATT:QueueItems'))     
        put(ind + 'DeleteQueueItems(')               
        self.display_name(element, ind2, 'Delete Queue Items')
        put(ind2 + 'Items: ' + queueitems)
        put(ind +')')

    # ------------ AddToCollection -------------
    @exception_handle_decorator
    def AddToCollection(self, element, ind, short, braced=False):
        ind2 = self.indent(ind)
        collection = self.show_arg(find(element, 'ATT:Collection'))
        item = self.show_arg(find(element, 'ATT:Item'))  
        put(ind + 'AddToCollection(')               
        self.display_name(element, ind2, 'Add To Collection')
        put(ind2 + 'Collection: ' + collection)
        put(ind2 + 'Item: ' + item)
        put(ind +')')     

    # ------------ ShouldStop -------------
    @exception_handle_decorator
    def ShouldStop(self, element, ind, short, braced=False):
        ind2 = self.indent(ind)
        result = self.show_arg(find(element, 'ATT:Result'))     
        put(ind + 'ShouldStop(')               
        self.display_name(element, ind2, 'Should Stop')
        put(ind2 + 'Out: ' + result)
        put(ind +')')

    # ------------ CloseWindow -------------
    @exception_handle_decorator
    def CloseWindow(self, element, ind, short, braced=False):
        ind2 = self.indent(ind)
        usewindow = self.show_arg(find(element, 'ATT:UseWindow'))     
        put(ind + 'CloseWindow(')    
        self.display_name(element, ind2, 'Close Window')
        put(ind2 + 'Window: ' + usewindow)
        put(ind +')')

    # ------------ GetRobotCredential --------
    @exception_handle_decorator
    def GetRobotCredential(self, element, ind, short, braced=False):
        ind2 = self.indent(ind)
        asset = find(element,'ATT:AssetName')
        asset = self.escape(asset)
        username = self.show_arg(find(element,'ATT:Username'))
        password = self.show_arg(find(element,'ATT:Password'))
        put(ind + 'GetRobotCredential(')    
        self.display_name(element, ind2, 'Get Robot Credential')
        put(ind2 + 'Asset:' + asset)
        put(ind2 + 'Username:' + username)
        put(ind2 + 'Password:' + password)       
        put(ind +')')

    # ------------ RetryScope --------
    @exception_handle_decorator
    def RetryScope(self, element, ind, short, braced=False):
        ind2 = self.indent(ind)
        number_of_retries = find(element,'ATT:NumberOfRetries')
  
        put(ind + 'RetryScope(')    
        self.display_name(element, ind2, 'Retry Scope')
        put(ind2 + 'Number of Retries:' + number_of_retries)
        put(ind2 + 'Condition: {')
        self.display_child(element, ind2, short, 'RetryScope.Condition','ActivityFunc')
        put(ind2 + 'Retry: {')
        self.display_variables(element, ind, short)
        self.display_child(element, ind2, short, 'Body','ActivityAction')
        put(ind + ')')

    # ------------ TypeSecureText --------------
    @exception_handle_decorator
    def TypeSecureText(self, element, ind, short, braced=False):
        ind2 = self.indent(ind)
        text = self.show_arg(find(element,'ATT:SecureText'))
        selector = find(element,'Target','Target','ATT:Selector')

        put(ind + 'TypeSecureText(')    
        self.display_name(element, ind2, 'Type Secure Text')
        put(ind2 + 'Text: ' + text)
        self.display_selector (selector, ind2)
        put(ind + ')')

    # ------------ TypeSecureText --------------
    @exception_handle_decorator
    def Check(self, element, ind, short, braced=False):
        ind2 = self.indent(ind)
        selector = find(element,'Target','Target','ATT:Selector')

        put(ind + 'Check(')    
        self.display_name(element, ind2, 'Check')
        self.display_selector (selector, ind2)
        put(ind + ')')
        
# Make some parser functions global to save typing
global_parser = parser()

def find(element, *targets):
    return global_parser.find(element, *targets)

def find_or_none(element, *targets):
    return global_parser.find_or_none(element, *targets)

def multifind(element, target):
    return global_parser.multifind(element, target)

def put(text):
    print(text)
    if out_file:
        out_file.write(text+'\n')

# Create and start the processor
def doit(in_file, short, hash_selector, show_datatable):
    data = ET.parse(in_file)     
    root = data.getroot()
    proc = processor()
    proc.start(root,'', short, hash_selector, show_datatable)


in_file = None
out_file = None
rethrow = False # Set true to debug this project

if __name__ == "__main__":
    
    # Parse -in commandline argument
    try:
        in_file = sys.argv[sys.argv.index('-in') + 1]
    except Exception:
        pass

    # Parse -in commandline argument     
    try:
        out_file = sys.argv[sys.argv.index('-out') + 1]
    except Exception:
        pass

    # Parse commandline flags
    hash_selector = '-hashselector' in sys.argv
    show_datatable = '-showdatatable' in sys.argv 
    short = '-short' in sys.argv 
      
    if in_file:
        if out_file:
            with open(out_file, 'w') as file:
                out_file = file  
                doit(in_file, short, hash_selector, show_datatable)
        else:
                doit(in_file, short, hash_selector, show_datatable)
                print('press enter to exit')
                sys.stdin.readline()
    else:
        print()
        print('Usage')
        print('=====')
        print('-in [filepathto.xaml]  (manditory)')
        print('-out [outputfilepath]  (optional) save the result as a file')
        print('-short                 (optional) causes some activities to be represented shorter')
        print('-hashselector          (optional) causes selectors to represented as a hashcode')
        print('-showdatatable         (optional) causes initialisation of datatable to be shown')
        time.sleep(1)
        print('press enter to exit')
        sys.stdin.readline()
